#!

$script = <<'EOD';
~startperl~ -w
use bytes;

# # # # # # # # # # # #

$pp_pos = 0;
@pp_closers = ();

sub pp_open {
  local($o, $f, $c) = @_;
  
  if ($pp_pos > 0) {
    $pp_pos++;
    print STDERR ' ';
    if (length($o.$f.$c) + $pp_pos > 79) {
      print STDERR "\n";
      $pp_pos = length (join '', @pp_closers);
    }
  }
  
  $pp_pos += length($o.$f.$c);
  print STDERR $o, $f;
  
  push(@pp_closers, $c);
}


sub pp_close {
  local($c) = pop(@pp_closers);
  print STDERR $c;
}


sub pp_end {
  print STDERR "\n" if ($pp_pos > 0);
  $pp_pos = 0;
}

# # # # # # # # # # # #

$RECREATE = 0;
$CVS = 0;

sub writepsres {
  my($filename) = @_;
  
  open(IDX, ">$filename");
  select IDX;
  print "PS-Resources-Exclusive-1.0\n";
  print "FontOutline\n" if %FontOutline;
  print "FontAFM\n" if %FontAFM;
  print "FontAMFM\n" if %FontAMFM;
  print "FontFamily\n" if %FontFamily;
  print "fontDatabases\n" if %fontDatabases;
  print "fontFoundry\n" if %fontFoundry;
  print ".\n";
  
  my($x, $y);
  if (%FontOutline) {
    print "FontOutline\n";
    foreach $x (sort keys %FontOutline) {
      print $x, '=', $FontOutline{$x}, "\n";
    }
    print ".\n";
  }
  if (%FontAFM) {
    print "FontAFM\n";
    foreach $x (sort keys %FontAFM) {
      print $x, '=', $FontAFM{$x}, "\n";
    }
    print ".\n";
  }
  if (%FontAMFM) {
    print "FontAMFM\n";
    foreach $x (sort keys %FontAMFM) {
      print $x, '=', $FontAMFM{$x}, "\n";
    }
    print ".\n";
  }
  if (%FontFamily) {
    print "FontFamily\n";
    foreach $x (sort keys %FontFamily) {
      print $x, $FontFamily{$x}, "\n";
    }
    print ".\n";
  }
  if (%fontDatabases) {
    print "fontDatabases\n";
    foreach $x (sort keys %fontDatabases) {
      print $x, '=', $fontDatabases{$x}, "\n";
    }
    print ".\n";
  }
  if (%fontFoundry) {
    print "fontFoundry\n";
    foreach $x (sort keys %fontFoundry) {
      print $x, '=', $fontFoundry{$x}, "\n";
    }
    print ".\n";
  }
  
  close(IDX);

  if ($CVS && $filename =~ /\A(.*\/|)PSres.upr\Z/) {
      my($cvsignore) = $1 . ".cvsignore";
      if (!(-r $cvsignore) && open(FOO, ">$cvsignore")) {
	  print FOO ".cvsignore PSres.upr\n";
	  close FOO;
      }
  }
}


sub find_suffix ($@) {
    my($base) = shift @_;
    while (@_) {
	my($suffix) = shift @_;
	return "$base.$suffix" if -f "$base.$suffix";
    }
    undef;
}

sub makepsres {
  
  my(%read_files, @files, $filename, $familyname, $fullname, $notice, $x);
  local(%FontOutline, %FontAFM, %FontAMFM, %FontFamily, %fontDatabases,
	%fontFoundry);
  my($afm, $AFM, $pfa, $pfb, $pfx, $amfm, $fam, $full, $fam2, $full2);
  
  opendir(DIR, ".");
  @files = grep(/\.(pfa|pfb|amfm|afm)$/i && -f, readdir(DIR));
  closedir(DIR);
  
  local($/);
  foreach $filename (@files) {
    next if $read_files{$filename};
    
    $filename =~ s/\.(pfa|pfb|amfm|afm)$//i;

    $afm = find_suffix($filename, 'afm', 'AFM');
    $amfm = find_suffix($filename, 'amfm', 'AMFM');
    $pfa = find_suffix($filename, 'pfa', 'PFA');
    $pfb = find_suffix($filename, 'pfb', 'PFB');
    
    $pfx = $pfb ? $pfb : $pfa;
    die if !$pfx && !$afm && !$amfm;
    
    $fontname = $fullname = $familyname = $notice = "";
    if ($afm || $amfm) {
      $/ = 'StartCharMetrics';
      open(AFM, ($afm ? $afm : $amfm));
      $_ = <AFM>;
      close AFM;
      ($fontname) = m|FontName \s+ (.*?) \s* [\r\n]|mx;
      ($fullname) = m|FullName \s+ (.*?) \s* [\r\n]|mx;
      ($familyname) = m|FamilyName \s+ (.*?) \s* [\r\n]|mx;
      ($notice) = m|Notice \s+ (.*?) \s* [\r\n]|mx;
      
    } elsif ($pfx) {
      $/ = 'eexec';
      open(PFX, $pfx);
      $_ = <PFX>;
      close(PFX);
      tr/\r\n/\n\n/s;
      ($fontname) = m|/FontName\s*/(\S+) def|;
      ($fullname) = m|/FullName\s*\(([^\)]+)\)|;
      ($familyname) = m|/FamilyName\s*\(([^\)]+)\)|;
      ($notice) = m#/Notice \s* \( ( ( [^\(\)] | \( [^\)]+ \) )+ ) \)#x;
    }

    next if !defined($fullname) || !defined($fontname) || !defined($familyname);
    
    $fontname =~ s/\s+$//;
    $fullname =~ s/\s+$//;
    $familyname =~ s/\s+$//;
    if ($fontname) {
      $FontOutline{$fontname} = $pfx if $pfx;
      $FontAFM{$fontname} = $afm if $afm;
      $FontAMFM{$fontname} = $amfm if $amfm;
      
      $fam = $familyname;
      $full = $fullname;

      # Want to force the family name to be a prefix of the full name.
      
      # Sometimes family names don't contain spaces, but should;
      # e.g. JansonText.
      if ($fam && $full && $full !~ /^$fam/) {
	  $fam2 = $fam;
	  $fam2 =~ s/([a-z0-9])([A-Z])/$1 $2/g;
	  $fam = $fam2 if $full =~ /^$fam2/;
      }

      # Sometimes the full name has only a prefix of the family name;
      # e.g. Helvetica Neue <-> Helvetica 45 Light.
      if ($fam && $full && $full !~ /^$fam/) {
	  $fam2 = $fam;
	  while ($fam2 =~ /^(\S*.*\S) (\S*)$/) {
	      $fam2 = $1;
	      if ($full =~ /^$fam2(?!\w)(.*)/) {
		  $full = $fam . $1;
		  last;
	      }
	  }
      }

      # Linotype full names can follow a weird pattern;
      # e.g. "12 Linotype Centennial* 55 Roman   05633".
      if ($fam && $full && $full !~ /^$fam/ && $full =~ /\s$fam\*\s*(.*?)\s+\d+$/) {
	  $full = $fam . " $1";
      }
      
      # Only get font family information from actual PFA/PFB fonts.
      if ($fam && $full) {
	  if ($full =~ /^$fam(?![a-z])(.*)$/) {
	      $alter = $1;
	  } else {
	      $alter = $full;
	  }
	  $alter = 'Plain' if $alter =~ /^\s*$/;
	  $alter =~ s/^[-\s]*//;
	  $alter =~ s/Oldstyle Figures/OsF/;
	  $alter =~ s/Small Caps( \& OsF)?/SC/;
	  
	  $_ = ($FontFamily{$familyname} || '');
	  if (!/[=,]$alter([=,]|$)/) {
	      $_ .= $_ ? ',' : '=';
	      $FontFamily{$familyname} = $_ . "$alter,$fontname";
	  }
      }
      
      while (($f, $n) = each %fontDatabasesRX) {
	$fontDatabases{$fontname} = $n if $fontname =~ /^$f$/;
      }

      $notice = '' if !defined($notice);
      $fontFoundry{$fontname} = 'Emigre' if ($notice =~ /Emigre/);
      $fontFoundry{$fontname} = 'Letraset' if ($notice =~ /LETRASET/i);
      $fontFoundry{$fontname} = 'Font Bureau' if ($notice =~ /Font Bureau/i);
      $fontFoundry{$fontname} = 'TeX' if ($notice =~ /American Mathematical/i);
      $fontFoundry{$fontname} = 'Bigelow & Holmes' if ($notice =~ /Bigelow & Holmes/i);
      $fontFoundry{$fontname} = 'URW' if ($notice =~ /\bURW/);
    }
  }
  
  if (%FontOutline || %FontAFM || %FontAMFM || %FontFamily || %fontDatabases ||
      %fontFoundry) {
    &writepsres("PSres.upr");
  } else {
    unlink("PSres.upr");
  }
  closedir(DIR);
}


sub readpsrescomp ($\@$) {
  my($DIR, $idxref, $type) = @_;
  
  if ($type =~ /FontOutline/) {
    while (defined($_ = shift @$idxref) && $_ ne '.') {
      $FontOutline{$1} = $DIR . $2 if (/^([^=]*)=(.*)$/);
    }
  } elsif ($type =~ /FontAFM/) {
    while (defined($_ = shift @$idxref) && $_ ne '.') {
      $FontAFM{$1} = $DIR . $2 if (/^([^=]*)=(.*)$/);
    }
  } elsif ($type =~ /FontAMFM/) {
    while (defined($_ = shift @$idxref) && $_ ne '.') {
      $FontAMFM{$1} = $DIR . $2 if (/^([^=]*)=(.*)$/);
    }
  } elsif ($type =~ /FontFamily/) {
    while (defined($_ = shift @$idxref) && $_ ne '.') {
      if (/^([^=]*)=(.*)$/) {
	$FontFamily{$1} .= ($FontFamily{$1} ? ',' : '=');
	$FontFamily{$1} .= $2;
      }
    }
  } elsif ($type =~ /fontDatabases/) {
    while (defined($_ = shift @$idxref) && $_ ne '.') {
      $fontDatabases{$1} = $2 if (/^([^=]*)=(.*)$/);
    }
  } elsif ($type =~ /fontFoundry/) {
    while (defined($_ = shift @$idxref) && $_ ne '.') {
      $fontFoundry{$1} = $2 if (/^([^=]*)=(.*)$/);
    }
  } else {
    while (defined($_ = shift @$idxref) && $_ ne '.') {
    }
  }
}


sub readpsres {
  my($file, $DIR) = @_;
  
  if ($file =~ m|^(.*/)[^/]*$|) {
    $DIR .= $1;
  }
  
  undef $/;
  open(IDX, $file);
  $_ = <IDX>;
  close(IDX);

  my(@idx);
  @idx = split /\s*[\r\n]+/;

  while (defined($_ = shift @idx) && $_ ne '.') {
  }
  $_ = shift @idx;
  if (m|^/(/.*)|) {
    $DIR = $1;
    $_ = shift @idx;
  } elsif (m|^/(.*)|) {
    $DIR .= "/$1";
    $_ = shift @idx;
  }
  
  readpsrescomp($DIR, @idx, $_);
  while (@idx) {
    readpsrescomp($DIR, @idx, shift @idx);
  }
  close IDX;
}


# index notation: filename fontname
sub readindex {
  my($directoryname, $DIR) = @_;
  my($closeit) = $directoryname ne '.'
      && $directoryname !~ /\.fm$/
      && $directoryname ne 'CVS';
  
  undef %fontDatabasesRX;
  
  if (-r ".fontscan") {
      undef $/;
      open(COMMANDS, ".fontscan");
      $_ = <COMMANDS>;
      close(COMMANDS);
      
      my(@cmd);
      @cmd = split /\s*[\r\n]+/;

      while (defined($_ = shift @cmd)) {
	  chomp;
	  next if /^\s*$/ || /^\s*[\#\%]/;
	  return 0 if (/^\s*ignore\b/);
	  $fontDatabasesRX{$1} = $2 if (m|^\s*fontdb\s+(\S+)\s+(.*)\s*$|);
      }
  }
  
  &pp_open('(', $directoryname, ')') if $closeit;
  
  if ($RECREATE) {
    &makepsres;
    
  } else {
    opendir(DIR, ".");
    @files = grep(/^[^.].*\.(pfa|pfb|afm|amfm)$/i && -f, readdir(DIR));
    closedir(DIR);
    
    $indextime = 0;
    $indextime = (stat(_))[9] if -s "PSres.upr";
    $indextime = 0 if ((stat("."))[9] > $indextime);
    $indextime = 0 if (-r ".fontscan" && (stat(".fontscan"))[9] > $indextime);
    foreach $file (@files) {
      if ((stat($file))[9] > $indextime) {
	&makepsres;
	last;
      }
    }
  }
  
  &readpsres("PSres.upr", $DIR) if -r "PSres.upr";
  return $closeit;
}


sub readindexes {
  my($DIR, $dirname) = @_;
  my(@dirs, $closeit);
  
  opendir(DIR, ".");
  @dirs = grep(!/^\./ && -d && !-l && !/^mac$/, readdir(DIR));
  closedir(DIR);
  
  $closeit = &readindex($dirname, $DIR);
  foreach $d (@dirs) {
    chdir($d);
    &readindexes("$DIR$d/", $d);
    chdir("..");
  }
  &pp_close if $closeit;
}


foreach $_ (@ARGV) {
  if (/^-r$/ || /^--recreate$/
      || /^-a$/ || /^--all$/) {
    $RECREATE = 1;
  } elsif (/^--cvs$/) {
    $CVS = 1;
  }
}

unlink('PSres.upr');
&readindexes('', '.');
&writepsres('PSres.upr');

&pp_end;

__END__

=head1 NAME

autofont-mkpsres - create a PSres.upr file for a font directory

=head1 SYNOPSIS

  autofont-mkpsres [-a]

=head1 DESCRIPTION

B<autofont-mkpsres> searches the current directory, and any subdirectories,
for PostScript fonts, and generates one or more F<PSres.upr> files
summarizing information about available fonts.

Normally B<autofont-mkpsres> uses file modification times to avoid
generating F<PSres.upr> files for subdirectories that have not changed. The
B<-a> flag prevents this behavior.

B<autofont-mkpsres> implements a subset of L<makepsres(1)>'s functionality,
but it handles that subset better in practice (it is better at recognizing
.pfb fonts, for example).

=head1 BUGS

Report bugs and send patches and enhancement requests to
C<ekohler@gmail.com>.

=head1 SEE ALSO

B<makepsres(1)>

Appendix A and Appendix B of "Display PostScript Toolkit for X" in
I<Programming the Display PostScript System with X>, for information on the
F<PSres.upr> format

=head1 AUTHOR

Eddie Kohler, C<ekohler@gmail.com>

See L<http://www.lcdf.org/type/> for updates.

EOD

require Config;
$cfg{'startperl'} = $Config::Config{'startperl'};
$cfg{'libdir'} = `make libdir.txt >/dev/null 2>&1 && cat libdir.txt`;
chomp $cfg{'libdir'};

$script =~ s/\~(\w+)\~/$cfg{$1}/eg;

if (!open(FILE, ">autofont-mkpsres") || !print FILE ($script) || !close(FILE)) {
    die "Error while writing script: $!";
}
